/*
 * Copyright © 2011 Canonical Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Canonical Ltd. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Canonical Ltd. makes no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * CANONICAL, LTD. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL CANONICAL, LTD. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authored by: Sam Spilsbury <sam.spilsbury@canonical.com>
 */

#include <tr1/tuple>
#include <test-constrain-to-workarea.h>
#include <constrain-to-workarea.h>
#include <iostream>
#include <stdlib.h>
#include <cstring>

namespace cw = compiz::window;
namespace cwe = cw::extents;
namespace cp = compiz::place;

using ::testing::WithParamInterface;
using ::testing::ValuesIn;
using ::testing::Combine;

class PlaceClampGeometryToWorkArea :
    public CompPlaceTest
{
    public:

	PlaceClampGeometryToWorkArea () :
	    screensize (1000, 2000),
	    workArea (50, 50, 900, 1900),
            flags (0)
        {
	    memset (&extents, 0, sizeof (cwe::Extents));
        }

    protected:

	CompSize screensize;
	CompRect workArea;
	cw::Geometry g;
	cwe::Extents extents;
	unsigned int flags;
};

namespace
{
    const cw::Geometry LimitOfAllowedGeometry (50, 50, 900, 1900, 0);
}

namespace compiz
{
namespace window
{
std::ostream & operator<<(std::ostream &os, const Geometry &g)
{
    return os << "Window Geometry: (" << g.x () <<
                 ", " << g.y () <<
                 ", " << g.width () <<
                 ", " << g.height () <<
                 ", " << g.border () << ")";
}
namespace extents
{
std::ostream & operator<<(std::ostream &os, const Extents &e)
{
    return os << "Window Extents: (left: " << e.left <<
                 ", right: " << e.right <<
                 ", top: " << e.top <<
                 ", bottom: " << e.bottom << ")";
}
}
}
}

std::ostream & operator<<(std::ostream &os, const CompPoint &p)
{
    return os << "Point: (" << p.x () << ", " << p.y () << ")";
}

TEST_F (PlaceClampGeometryToWorkArea, NoConstrainmentRequired)
{
    g = cw::Geometry (100, 100, 200, 200, 0);

    /* Do nothing */
    cp::clampGeometryToWorkArea (g, workArea, extents, flags, screensize);

    EXPECT_EQ (g, cw::Geometry (100, 100, 200, 200, 0));
}

TEST_F (PlaceClampGeometryToWorkArea, LargerThanWorkAreaConstrainsToWorkAreaSize)
{
    /* Larger than workArea */
    g = cw::Geometry (50, 50, 950, 1950, 0);

    cp::clampGeometryToWorkArea (g, workArea, extents, flags, screensize);

    EXPECT_EQ (g, LimitOfAllowedGeometry);
}

TEST_F (PlaceClampGeometryToWorkArea, OutsideTopLeftConstrainment)
{
    /* Outside top left */
    g = cw::Geometry (0, 0, 900, 1900, 0);

    cp::clampGeometryToWorkArea (g, workArea, extents, flags, screensize);

    EXPECT_EQ (g, LimitOfAllowedGeometry);
}

TEST_F (PlaceClampGeometryToWorkArea, OutsideTopRightConstrainment)
{
    /* Outside top right */
    g = cw::Geometry (100, 0, 900, 1900, 0);

    cp::clampGeometryToWorkArea (g, workArea, extents, flags, screensize);

    EXPECT_EQ (g, LimitOfAllowedGeometry);
}

TEST_F (PlaceClampGeometryToWorkArea, OutsideBottomLeftConstrainment)
{
    /* Outside bottom left */
    g = cw::Geometry (0, 100, 900, 1900, 0);

    cp::clampGeometryToWorkArea (g, workArea, extents, flags, screensize);

    EXPECT_EQ (g, LimitOfAllowedGeometry);
}

TEST_F (PlaceClampGeometryToWorkArea, OutsideBottomRightConstrainment)
{
    /* Outside bottom right */
    g = cw::Geometry (100, 100, 900, 1900, 0);

    cp::clampGeometryToWorkArea (g, workArea, extents, flags, screensize);

    EXPECT_EQ (g, LimitOfAllowedGeometry);
}

TEST_F (PlaceClampGeometryToWorkArea, NoChangePositionIfSizeUnchanged)
{
    /* For the size only case, we should not
     * change the position of the window if
     * the size does not change */
    g = cw::Geometry (0, 0, 900, 1900, 0);
    flags = cp::clampGeometrySizeOnly;

    cp::clampGeometryToWorkArea (g, workArea, extents, flags, screensize);

    EXPECT_EQ (g, cw::Geometry (0, 0, 900, 1900, 0));
}

TEST_F (PlaceClampGeometryToWorkArea, ChangePositionIfSizeChanged)
{
    g = cw::Geometry (0, 0, 1000, 2000, 0);
    flags = cp::clampGeometrySizeOnly;

    cp::clampGeometryToWorkArea (g, workArea, extents, flags, screensize);

    EXPECT_EQ (g, LimitOfAllowedGeometry);
}

namespace
{
typedef std::tr1::tuple <cw::Geometry, cwe::Extents> ConstrainPositionToWorkAreaParam;

const cw::Geometry & WindowGeometry (const ConstrainPositionToWorkAreaParam &p)
{
    return std::tr1::get <0> (p);
}

const cwe::Extents & WindowExtents (const ConstrainPositionToWorkAreaParam &p)
{
    return std::tr1::get <1> (p);
}

CompPoint InitialPosition (const ConstrainPositionToWorkAreaParam &p)
{
    /* Initial position is where the window is right now */
    return (std::tr1::get <0> (p)).pos ();
}

const CompRect  WArea (50, 50, 900, 1900);
const CompPoint ExpectedPosition (WArea.pos ());
}

class PlaceConstrainPositionToWorkArea :
    public CompPlaceTest,
    public WithParamInterface <ConstrainPositionToWorkAreaParam>
{
    public:

	PlaceConstrainPositionToWorkArea ()
        {
	    memset (&extents, 0, sizeof (cwe::Extents));
        }

    protected:

	CompRect workArea;
	cw::Geometry g;
	cwe::Extents extents;
};

TEST_P (PlaceConstrainPositionToWorkArea, PositionConstrainedWithExtents)
{
    g = WindowGeometry (GetParam ());
    extents = WindowExtents (GetParam ());

    CompPoint pos = InitialPosition (GetParam ());
    pos = cp::constrainPositionToWorkArea (pos, g, extents, WArea, false);

    const CompPoint expectedAfterExtentsAdjustment = ExpectedPosition +
						     CompPoint (extents.left,
								extents.top);

    EXPECT_EQ (expectedAfterExtentsAdjustment, pos);
}

TEST_P (PlaceConstrainPositionToWorkArea, PositionConstrainedStaticGravity)
{
    g = WindowGeometry (GetParam ());
    extents = WindowExtents (GetParam ());

    CompPoint pos = InitialPosition (GetParam ());
    pos = cp::constrainPositionToWorkArea (pos, g, extents, WArea, true);

    /* Do not adjust residual position for extents with windows
     * that have a static gravity */
    EXPECT_EQ (ExpectedPosition, pos);
}

namespace
{
cwe::Extents PossibleExtents[] =
{
    cwe::Extents (0, 0, 0, 0),
    cwe::Extents (1, 0, 0, 0),
    cwe::Extents (0, 1, 0, 0),
    cwe::Extents (0, 0, 1, 0),
    cwe::Extents (0, 0, 0, 1)
};

cw::Geometry PossibleGeometries[] =
{
    cw::Geometry (WArea.x (), WArea.y (),
		  WArea.width (), WArea.height (), 0),
    cw::Geometry (WArea.x () - 1, WArea.y (),
		  WArea.width (), WArea.height (), 0),
    cw::Geometry (WArea.x (), WArea.y () - 1,
		  WArea.width (), WArea.height (), 0),
    cw::Geometry (WArea.x () + 1, WArea.y (),
		  WArea.width (), WArea.height (), 0),
    cw::Geometry (WArea.x (), WArea.y () + 1,
		  WArea.width (), WArea.height (), 0)
};
}

INSTANTIATE_TEST_CASE_P (PlacementData,
                         PlaceConstrainPositionToWorkArea,
                         Combine (ValuesIn (PossibleGeometries),
				  ValuesIn (PossibleExtents)));
